/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.interceptor;

import static java.lang.annotation.ElementType.*;

import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Map;

import org.t2framework.commons.aop.spi.Interceptor;
import org.t2framework.commons.aop.spi.Invocation;
import org.t2framework.commons.util.CollectionsUtil;
import org.t2framework.commons.util.Reflections.GenericsUtil;

/**
 * 
 * {@.en Exception handle interceptor.}
 * 
 * <br />
 * 
 * {@.ja 例外処理用の{@link Interceptor}です.継承して、パラメタライズされた型Tを適切な例外にして使用します.}
 * 
 * @author shot
 * 
 * @param <T>
 *            exception that should handle this interceptor.
 */
public abstract class ThrowsInterceptor<T extends Throwable> implements
		Interceptor {

	/**
	 * 
	 * {@.en }
	 * 
	 * <br />
	 * 
	 * {@.ja 例外ハンドル用カスタマイズアノテーションです.
	 * {@link ThrowsInterceptor#handle(Throwable, Invocation)}
	 * よりも優先して呼ばせたいメソッドが出てきた場合、このアノテーションをつけたメソッドを準備しておくと、適切にコールされます.}
	 * 
	 * @author shot
	 * 
	 */
	@Inherited
	@Retention(RetentionPolicy.RUNTIME)
	@Target( { METHOD })
	public static @interface HandleThrowable {
	}

	protected Map<Class<?>, Method> methodMap = CollectionsUtil.newHashMap();

	protected final Class<?> exceptionType;

	/**
	 * {@.en Constructs ThrowsInterceptor.}
	 * 
	 * <br />
	 * 
	 * {@.ja {@link ThrowsInterceptor}を構築します。}
	 */
	public ThrowsInterceptor() {
		Method[] methods = getClass().getMethods();
		for (int i = 0; i < methods.length; ++i) {
			Method m = methods[i];
			if (m.isBridge() || m.isSynthetic()) {
				continue;
			}
			if (isHandleThrowable(m)) {
				methodMap.put(m.getParameterTypes()[0], m);
			}
		}
		this.exceptionType = getExceptionType();
	}

	protected boolean isHandleThrowable(Method m) {
		final HandleThrowable annotation = m
				.getAnnotation(HandleThrowable.class);
		if (annotation == null) {
			return false;
		}
		final Class<?>[] paramTypes = m.getParameterTypes();
		if (paramTypes.length == 2
				&& Throwable.class.isAssignableFrom(paramTypes[0])
				&& Invocation.class.isAssignableFrom(paramTypes[1])) {
			return true;
		} else {
			throw singnatureMismatch(m);
		}
	}

	protected RuntimeException singnatureMismatch(Method m) {
		return new IllegalStateException("Signature unmatch:"
				+ m.toGenericString());
	}

	@SuppressWarnings("unchecked")
	@Override
	public Object intercept(Invocation<Method> invocation) throws Throwable {
		try {
			return invocation.proceed();
		} catch (Throwable t) {
			Method method = getMethod(t);
			if (method != null) {
				try {
					return method.invoke(this, new Object[] { t, invocation });
				} catch (InvocationTargetException ite) {
					throw ite.getTargetException();
				}
			}
			if (this.exceptionType.isAssignableFrom(t.getClass())) {
				return handle((T) t, invocation);
			}
			throw t;
		}
	}

	protected Class<?> getExceptionType() {
		Class<?> clazz = getClass();
		Type type = null;
		do {
			type = clazz.getGenericSuperclass();
			clazz = clazz.getSuperclass();
		} while (clazz != ThrowsInterceptor.class);
		final Type actualType = ((ParameterizedType) type)
				.getActualTypeArguments()[0];
		final Class<?> exceptionClass = GenericsUtil.getRawClass(actualType);
		return exceptionClass;
	}

	protected abstract Object handle(T exception, Invocation<Method> invocation)
			throws T;

	private Method getMethod(Throwable t) {
		Class<?> clazz = t.getClass();
		Method method = methodMap.get(clazz);
		while (method == null && clazz.equals(Throwable.class) == false) {
			clazz = clazz.getSuperclass();
			method = methodMap.get(clazz);
		}
		return method;
	}
}
